import json
import subprocess as sub
from pathlib import Path
import signal
import sys
from typing import List
import argparse

import pytest
import yaml
from networkx.algorithms import isomorphism
from mergedeep import merge, Strategy

import sophios.cli
import sophios.compiler
import sophios.inlineing
import sophios.input_output
import sophios.run_local
import sophios.utils
import sophios.ast
import sophios.plugins
from sophios import auto_gen_header
from sophios.cli import get_args
from sophios.utils_yaml import wic_loader
from sophios.wic_types import NodeData, StepId, Yaml, YamlTree, Json
from sophios.utils_graphs import get_graph_reps

from .test_setup import tools_cwl, yml_paths, validator, yml_paths_tuples

# Look in each directory of search_paths_wic tag in global_config.json
# for separate config_ci.json files and combine them.
config_ci: Json = {}
global_config = sophios.input_output.get_config(Path(get_args().config_file), Path(get_args().config_file))
search_paths_wic_tag = global_config['search_paths_wic']
for _yml_namespaces in search_paths_wic_tag:
    yml_dirs = search_paths_wic_tag[_yml_namespaces]
    for yml_dir in yml_dirs:
        config_ci_json = Path(yml_dir) / 'config_ci.json'
        if config_ci_json.exists():
            print(f'Reading {config_ci_json}')
            with open(config_ci_json) as f:
                contents = f.read().splitlines()
                # Strip out comments. (Comments are not allowed in JSON)
                contents = [line for line in contents if not line.strip().startswith('//')]
                config_ci_tmp = json.loads('\n'.join(contents))
            # Use the Additive Strategy to e.g. concatenate lists
            config_ci = merge(config_ci, config_ci_tmp, strategy=Strategy.TYPESAFE_ADDITIVE)

# Due to the computational complexity of the graph isomorphism problem, we
# need to manually exclude large workflows.
# See https://en.wikipedia.org/wiki/Graph_isomorphism_problem
large_workflows: List[str] = config_ci.get("large_workflows", [])
yml_paths_tuples_not_large = [(s, p) for (s, p) in yml_paths_tuples if s not in large_workflows]

# NOTE: Most of the workflows in this list have free variables because they are subworkflows
# i.e. if you try to run them, you will get "Missing required input parameter"
run_blacklist: List[str] = config_ci.get("run_blacklist", [])
run_weekly: List[str] = config_ci.get("run_weekly", [])
run_partial_failures: List[str] = config_ci.get("run_partial_failures", [])

yml_paths_tuples_weekly = [(s, p) for (s, p) in yml_paths_tuples if s in run_weekly]

yml_paths_tuples_not_blacklist_on_push = [(s, p) for (s, p) in yml_paths_tuples
                                          if s not in run_blacklist and s not in run_weekly]

yml_paths_partial_failure = [(s, p) for (s, p) in yml_paths_tuples
                             if s in run_partial_failures]

# currently [vs_demo_2, vs_demo_3, vs_demo_4, elm, nmr,
#            multistep1, multistep2, multistep3, helloworld, scattering_scaling]


def is_isomorphic_with_timeout(g_m: isomorphism.GraphMatcher, yml_path_str: str) -> None:
    """Calls the .is_isomorphic() method with a timeout of 10 seconds.

    Args:
        gm (isomorphism.GraphMatcher): The graph isomorphism object.
        yml_path_str (str): The root yml workflow file for error reporting.
    """
    def handler(_signum, _frame):  # type: ignore
        line1 = 'Graph isomorphism check timed out after 10 seconds.'
        line2 = f'Consider adding {yml_path_str} to the large_workflows list.'
        raise Exception(f'{line1}\n{line2}')

    # See https://github.com/bokeh/bokeh/issues/11627#issuecomment-921576787
    if sys.platform == 'win32':
        # Windows does not support alarm, so just use a regular call here.
        # Note that there is a backup timeout in the github action workflow
        assert g_m.is_isomorphic()  # See top-level comment above!
    else:
        # See https://docs.python.org/3/library/signal.html#examples
        # NOTE: You CANNOT use `pytest --workers 8 ...` with this. Otherwise:
        # "ValueError: signal only works in main thread of the main interpreter"
        signal.signal(signal.SIGALRM, handler)

        signal.alarm(10)  # timeout after 10 seconds
        assert g_m.is_isomorphic()  # See top-level comment above!
        signal.alarm(0)  # Disable the alarm


@pytest.mark.slow
@pytest.mark.parametrize("yml_path_str, yml_path", yml_paths_tuples_not_blacklist_on_push)
def test_run_workflows_on_push(yml_path_str: str, yml_path: Path, cwl_runner: str) -> None:
    """Runs all of the workflows auto-discovered from the various
       directories in 'search_paths_wic', excluding all workflows which have been
       blacklisted in the various config_ci.json files and excluding the weekly
       workflows."""
    args = get_args(str(yml_path))
    run_workflows(yml_path_str, yml_path, cwl_runner, args)


@pytest.mark.slow
@pytest.mark.parametrize("yml_path_str, yml_path", yml_paths_tuples_not_blacklist_on_push)
def test_run_inlined_workflows_on_push(yml_path_str: str, yml_path: Path, cwl_runner: str) -> None:
    """Inlines and runs all of the workflows auto-discovered from the various
       directories in 'search_paths_wic', excluding all workflows which have been
       blacklisted in the various config_ci.json files and excluding the weekly
       workflows."""
    args = get_args(str(yml_path), ['--cwl_inline_subworkflows'])
    run_workflows(yml_path_str, yml_path, cwl_runner, args)


# partial failure tests
@pytest.mark.skip_pypi_ci
@pytest.mark.parametrize("yml_path_str, yml_path", yml_paths_partial_failure)
def test_run_partial_failures_pass(yml_path_str: str, yml_path: Path, cwl_runner: str) -> None:
    """Run workflows allowing partial failures. yml files of workflows which are known to have failure steps"""
    args = get_args(str(yml_path), ['--partial_failure_enable'])
    run_workflows(yml_path_str, yml_path, cwl_runner, args)


@pytest.mark.xfail()
@pytest.mark.parametrize("yml_path_str, yml_path", yml_paths_partial_failure)
def test_run_partial_failures_xfail(yml_path_str: str, yml_path: Path, cwl_runner: str) -> None:
    """Run workflows with known failures but without partial failure cli flag. It is expected to fail"""
    args = get_args(str(yml_path), [])
    run_workflows(yml_path_str, yml_path, cwl_runner, args)


@pytest.mark.slow
@pytest.mark.parametrize("yml_path_str, yml_path", yml_paths_tuples_weekly)
def test_run_workflows_weekly(yml_path_str: str, yml_path: Path, cwl_runner: str) -> None:
    """Runs all of the run_weekly workflows whitelisted in the various config_ci.json files."""
    args = get_args(str(yml_path))
    run_workflows(yml_path_str, yml_path, cwl_runner, args)


@pytest.mark.slow
@pytest.mark.parametrize("yml_path_str, yml_path", yml_paths_tuples_weekly)
def test_run_inlined_workflows_weekly(yml_path_str: str, yml_path: Path, cwl_runner: str) -> None:
    """Inlines and runs all of the run_weekly workflows whitelisted in the various config_ci.json files."""
    args = get_args(str(yml_path), ['--cwl_inline_subworkflows'])
    run_workflows(yml_path_str, yml_path, cwl_runner, args)


@pytest.mark.parametrize("yml_path_str, yml_path", yml_paths_tuples_not_blacklist_on_push)
def test_cwl_docker_extract(yml_path_str: str, yml_path: Path) -> None:
    """ Uses cwl-docker-extract to recursively `docker pull`"""
    args = get_args(str(yml_path))
    run_workflows(yml_path_str, yml_path, 'cwltool', args, True)
    return


def run_workflows(yml_path_str: str, yml_path: Path, cwl_runner: str, args: argparse.Namespace, docker_pull_only: bool = False) -> None:
    """Runs all of the given workflows."""

    # First compile the workflow.
    # Load the high-level yaml workflow file.
    with open(yml_path, mode='r', encoding='utf-8') as y:
        root_yaml_tree: Yaml = yaml.load(y.read(), Loader=wic_loader())
    Path('autogenerated/').mkdir(parents=True, exist_ok=True)
    wic_tag = {'wic': root_yaml_tree.get('wic', {})}
    plugin_ns = wic_tag['wic'].get('namespace', 'global')
    step_id = StepId(yml_path_str, plugin_ns)
    y_t = YamlTree(step_id, root_yaml_tree)
    yaml_tree_raw = sophios.ast.read_ast_from_disk(args.homedir, y_t, yml_paths, tools_cwl, validator,
                                                   args.ignore_validation_errors)
    with open(f'autogenerated/{Path(yml_path).stem}_tree_raw.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree_raw.yml))
    yaml_tree = sophios.ast.merge_yml_trees(yaml_tree_raw, {}, tools_cwl)
    with open(f'autogenerated/{Path(yml_path).stem}_tree_merged.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree.yml))
    root_yml_dir_abs = Path(args.yaml).parent.absolute()
    yaml_tree = sophios.ast.python_script_generate_cwl(yaml_tree, root_yml_dir_abs, tools_cwl)
    with open(f'autogenerated/{Path(yml_path).stem}_tree_python_script.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree.yml))

    if args.cwl_inline_subworkflows:
        while True:
            # Inlineing changes the namespaces, so we have to get new namespaces after each inlineing operation.
            namespaces_list = sophios.inlineing.get_inlineable_subworkflows(yaml_tree, tools_cwl, False, [])
            if namespaces_list == []:
                break

            yaml_tree, _len_substeps = sophios.inlineing.inline_subworkflow(yaml_tree, namespaces_list[0])

        # Append _inline here instead of in input_output.write_to_disk()
        step_id = StepId(yaml_tree.step_id.stem + '_inline', yaml_tree.step_id.plugin_ns)
        yaml_tree = YamlTree(step_id, yaml_tree.yml)

        with open(f'autogenerated/{Path(yml_path).stem}_tree_merged_inlined.wic', mode='w', encoding='utf-8') as f:
            f.write(yaml.dump(yaml_tree.yml))

    graph = get_graph_reps(str(yml_path))
    compiler_info = sophios.compiler.compile_workflow(yaml_tree, args, [], [graph], {}, {}, {}, {},
                                                      tools_cwl, True, relative_run_path=True, testing=True)
    rose_tree = compiler_info.rose
    sub_node_data: NodeData = rose_tree.data
    yaml_stem = sub_node_data.name

    rose_tree = sophios.plugins.cwl_prepend_dockerFile_include_path_rosetree(rose_tree)
    sophios.input_output.write_to_disk(rose_tree, Path('autogenerated/'), True, args.inputs_file)

    if docker_pull_only:
        # cwl-docker-extract recursively `docker pull`s all images in all subworkflows.
        # This is important because cwltool only uses `docker run` when executing
        # workflows, and if there is a local image available,
        # `docker run` will NOT query the remote repository for the latest image!
        # cwltool has a --force-docker-pull option, but this may cause multiple pulls in parallel.
        if args.container_engine == 'singularity':
            cmd = ['cwl-docker-extract', f'-s --dir {args.homedir}', f'autogenerated/{Path(yml_path).stem}.cwl']
        else:
            cmd = ['cwl-docker-extract', '--force-download', f'autogenerated/{Path(yml_path).stem}.cwl']
        sub.run(cmd, check=True)

        return

    if args.docker_remove_entrypoints:
        # Requires root, so guard behind CLI option
        if args.container_engine == 'docker':
            sophios.plugins.remove_entrypoints_docker()
        if args.container_engine == 'podman':
            sophios.plugins.remove_entrypoints_podman()

        rose_tree = sophios.plugins.dockerPull_append_noentrypoint_rosetree(rose_tree)
        sophios.input_output.write_to_disk(rose_tree, Path('autogenerated/'), True, args.inputs_file)

    if args.partial_failure_enable:
        rose_tree = sophios.plugins.cwl_update_outputs_optional_rosetree(rose_tree)
        sophios.input_output.write_to_disk(rose_tree, Path('autogenerated/'), True, args.inputs_file)
    # NOTE: Do not use --cachedir; we want to actually test everything.
    retval = sophios.run_local.run_local(args, rose_tree, None, cwl_runner, True)
    assert retval == 0

    # Finally, since there is an output file copying bug in cwltool,
    # we need to copy the output files manually. See comment above.
    if args.copy_output_files:
        sophios.run_local.copy_output_files(yaml_stem)


@pytest.mark.fast
@pytest.mark.serial
@pytest.mark.parametrize("yml_path_str, yml_path", yml_paths_tuples_not_large)
def test_cwl_embedding_independence(yml_path_str: str, yml_path: Path) -> None:
    """Tests that compiling a subworkflow is independent of how it is embedded
    into a parent workflow. Specifically, this compiles the root workflow and
    re-compiles every subworkflow (individually) as if it were a root workflow,
    then checks that the CWL for each subworkflow remains identical and checks
    that the embedded subworkflow DAGs and the re-compiled DAGs are isomorphic.
    """
    args = get_args(str(yml_path))

    # Load the high-level yaml workflow file.
    with open(yml_path, mode='r', encoding='utf-8') as y:
        root_yaml_tree: Yaml = yaml.load(y.read(), Loader=wic_loader())
    # Write the combined workflow (with all subworkflows as children) to disk.
    Path('autogenerated/').mkdir(parents=True, exist_ok=True)
    wic_tag = {'wic': root_yaml_tree.get('wic', {})}
    plugin_ns = wic_tag['wic'].get('namespace', 'global')
    step_id = StepId(yml_path_str + '.wic', plugin_ns)
    y_t = YamlTree(step_id, root_yaml_tree)
    yaml_tree_raw = sophios.ast.read_ast_from_disk(args.homedir, y_t, yml_paths, tools_cwl, validator,
                                                   args.ignore_validation_errors)
    with open(f'autogenerated/{yml_path.stem}_tree_raw.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree_raw.yml))
    yaml_tree = sophios.ast.merge_yml_trees(yaml_tree_raw, {}, tools_cwl)
    with open(f'autogenerated/{yml_path.stem}_tree_merged.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree.yml))
    root_yml_dir_abs = Path(args.yaml).parent.absolute()
    yaml_tree = sophios.ast.python_script_generate_cwl(yaml_tree, root_yml_dir_abs, tools_cwl)
    with open(f'autogenerated/{Path(yml_path).stem}_tree_python_script.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree.yml))

    # NOTE: The entire purpose of parsing an entire yaml forest is so we
    # can easily access the subtrees here. (i.e. without re-walking the AST)
    yaml_forest = sophios.ast.tree_to_forest(yaml_tree, tools_cwl)
    yaml_forest_lst = sophios.utils.flatten_forest(yaml_forest)
    yaml_forest_lst = [yf for yf in yaml_forest_lst if ".wic" not in yf.yaml_tree.step_id.stem]

    graph = get_graph_reps(str(yml_path))
    is_root = True
    compiler_info = sophios.compiler.compile_workflow(yaml_tree, args, [], [graph], {}, {}, {}, {},
                                                      tools_cwl, is_root, relative_run_path=False, testing=True)
    rose_tree = compiler_info.rose
    node_data_lst: List[NodeData] = sophios.utils.flatten_rose_tree(rose_tree)
    node_data_lst = [node for node in node_data_lst if ".wic" in node.name[0]]

    # This test doesn't necessarily need to write to disk, but useful for debugging.
    sophios.input_output.write_to_disk(rose_tree, Path('autogenerated/'), False, args.inputs_file)

    # Now, for each subworkflow of the given root workflow, compile the
    # subworkflow again from scratch, as if it were the root workflow,
    # and check that the generated CWL is identical. In other words,
    # check that the generated CWL of a subworkflow is independent of its
    # embedding into a parent workflow.
    assert len(node_data_lst[1:]) == len(yaml_forest_lst)  # There is an additional node
    for sub_node_data, sub_yaml_forest in zip(node_data_lst[1:], yaml_forest_lst):
        sub_name = sub_node_data.name
        assert sub_yaml_forest.yaml_tree.step_id.stem == sub_name + '.wic'

        # NOTE: Do we want to also test embedding independence with args.graph_inline_depth?
        # If so, we will need to patch testargs depending on len(sub_node_data.namespaces)
        # (due to the various instances of `if len(namespaces) < args.graph_inline_depth`)

        graph_fakeroot = get_graph_reps(str(sub_name))
        fake_root = True
        compiler_info_fakeroot = sophios.compiler.compile_workflow(sub_yaml_forest.yaml_tree, get_args(str(yml_path)),
                                                                   [], [graph_fakeroot], {}, {}, {}, {}, tools_cwl, fake_root, relative_run_path=False, testing=True)
        sub_node_data_fakeroot: NodeData = compiler_info_fakeroot.rose.data
        sub_cwl_fakeroot = sub_node_data_fakeroot.compiled_cwl

        # NOTE: Relative run: paths cause this test to fail, so remove them.
        # Using namespaced filenames in a single flat directory also
        # doesn't work because the namespaces will be of different lengths.
        sub_cwl_embedded = sophios.utils.recursively_delete_dict_key('run', sub_node_data.compiled_cwl)
        sub_cwl_fakeroot = sophios.utils.recursively_delete_dict_key('run', sub_cwl_fakeroot)

        if sub_cwl_embedded != sub_cwl_fakeroot:
            # Before we crash and burn, write out files for debugging.
            with open(f'{sub_name}_forest_embedded.wic', mode='w', encoding='utf-8') as w:
                w.write(yaml.dump(yaml_forest))
            with open(f'{sub_name}_forest_fakeroot.wic', mode='w', encoding='utf-8') as w:
                w.write(yaml.dump(sub_yaml_forest))
            # NOTE: Use _dot_cwl so we don't glob these files in get_tools_cwl()
            yaml_content = yaml.dump(sub_cwl_embedded, sort_keys=False, line_break='\n', indent=2)
            filename_emb = f'{sub_name}_embedded_dot_cwl'
            with open(filename_emb, mode='w', encoding='utf-8') as w:
                w.write('#!/usr/bin/env cwl-runner\n')
                w.write(auto_gen_header)
                w.write(''.join(yaml_content))
            yaml_content = yaml.dump(sub_cwl_fakeroot, sort_keys=False, line_break='\n', indent=2)
            filename_fake = f'{sub_name}_fakeroot_dot_cwl'
            with open(filename_fake, mode='w', encoding='utf-8') as w:
                w.write('#!/usr/bin/env cwl-runner\n')
                w.write(auto_gen_header)
                w.write(''.join(yaml_content))
            cmd = f'diff {filename_emb} {filename_fake} > {sub_name}.diff'
            sub.run(cmd, shell=True, check=False)
            print(f'Error! Check {filename_emb} and {filename_fake} and {sub_name}.diff')
        assert sub_cwl_embedded == sub_cwl_fakeroot

        # Check that the subgraphs are isomorphic.
        sub_graph_nx = sub_node_data.graph.networkx
        sub_graph_fakeroot_nx = sub_node_data_fakeroot.graph.networkx
        # assert isomorphism.faster_could_be_isomorphic(sub_graph_nx, sub_graph_fakeroot_nx)
        g_m = isomorphism.GraphMatcher(sub_graph_nx, sub_graph_fakeroot_nx)
        print('is_isomorphic()?', yml_path_str, sub_name)
        is_isomorphic_with_timeout(g_m, yml_path_str)


@pytest.mark.serial
@pytest.mark.parametrize("yml_path_str, yml_path", yml_paths_tuples_not_large)
def test_inline_subworkflows(yml_path_str: str, yml_path: Path) -> None:
    """Tests that compiling a workflow is independent of how subworkflows are inlined.
    Specifically, this inlines every subworkflow (individually) and checks that
    the original DAG and the inlined DAGs are isomorphic.
    """
    args = get_args(str(yml_path))
    # Load the high-level yaml workflow file.
    with open(yml_path, mode='r', encoding='utf-8') as y:
        root_yaml_tree: Yaml = yaml.load(y.read(), Loader=wic_loader())
    Path('autogenerated/').mkdir(parents=True, exist_ok=True)
    wic_tag = {'wic': root_yaml_tree.get('wic', {})}
    plugin_ns = wic_tag['wic'].get('namespace', 'global')
    step_id = StepId(yml_path_str, plugin_ns)
    y_t = YamlTree(step_id, root_yaml_tree)
    yaml_tree_raw = sophios.ast.read_ast_from_disk(args.homedir, y_t, yml_paths, tools_cwl, validator,
                                                   args.ignore_validation_errors)
    with open(f'autogenerated/{Path(yml_path).stem}_tree_raw.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree_raw.yml))
    yaml_tree = sophios.ast.merge_yml_trees(yaml_tree_raw, {}, tools_cwl)
    with open(f'autogenerated/{Path(yml_path).stem}_tree_merged.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree.yml))
    root_yml_dir_abs = Path(args.yaml).parent.absolute()
    yaml_tree = sophios.ast.python_script_generate_cwl(yaml_tree, root_yml_dir_abs, tools_cwl)
    with open(f'autogenerated/{Path(yml_path).stem}_tree_python_script.wic', mode='w', encoding='utf-8') as f:
        f.write(yaml.dump(yaml_tree.yml))

    namespaces_list = sophios.inlineing.get_inlineable_subworkflows(
        yaml_tree, tools_cwl, 'implementation' in wic_tag, [])
    if namespaces_list == []:
        assert True  # There's nothing to test

    graph = get_graph_reps(str(yml_path))
    compiler_info = sophios.compiler.compile_workflow(yaml_tree, args, [], [graph], {}, {}, {}, {},
                                                      tools_cwl, True, relative_run_path=True, testing=True)
    rose_tree = compiler_info.rose
    sub_node_data: NodeData = rose_tree.data

    sophios.input_output.write_to_disk(rose_tree, Path('autogenerated/'), True, args.inputs_file)

    # Inline each subworkflow individually and check that the graphs are isomorphic.
    for namespaces in namespaces_list:
        inline_yaml_tree, _len_substeps = sophios.inlineing.inline_subworkflow(yaml_tree, namespaces)

        inline_graph = get_graph_reps(str(yml_path))
        inline_compiler_info = sophios.compiler.compile_workflow(inline_yaml_tree, args,
                                                                 [], [inline_graph], {}, {}, {}, {}, tools_cwl, True, relative_run_path=True, testing=True)
        inline_rose_tree = inline_compiler_info.rose
        inline_sub_node_data: NodeData = inline_rose_tree.data

        # Check that the subgraphs are isomorphic.
        sub_graph_nx = sub_node_data.graph.networkx
        sub_graph_fakeroot_nx = inline_sub_node_data.graph.networkx
        # assert isomorphism.faster_could_be_isomorphic(sub_graph_nx, sub_graph_fakeroot_nx)
        g_m = isomorphism.GraphMatcher(sub_graph_nx, sub_graph_fakeroot_nx)
        print('is_isomorphic()?', yml_path_str, namespaces)
        is_isomorphic_with_timeout(g_m, yml_path_str)



import copy
from pathlib import Path
import re
from typing import Dict, List, Tuple

from mergedeep import merge, Strategy
import yaml

from . import utils
from .wic_types import Namespaces, Yaml, Tools, YamlTree, StepId, NodeData, RoseTree

# NOTE: AST = Abstract Syntax Tree

# TODO: Check for inline-ing subworkflows more than once and, if there are not
# any modifications from any parent dsl args, use yaml anchors and aliases.
# That way, we should be able to serialize back to disk without duplication.


def get_inlineable_subworkflows(yaml_tree_tuple: YamlTree,
                                tools: Tools,
                                implementation: bool = False,
                                namespaces_init: Namespaces = []) -> List[Namespaces]:
    """Traverses a yml AST and finds all subworkflows which can be inlined into their parent workflow.

    Args:
        yaml_tree_tuple (YamlTree): A tuple of name and yml AST
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()
        implementation (bool): True if the immediate parent workflow is a implementation.
        namespaces_init (Namespaces): The initial subworkflow to start the traversal ([] == root)

    Returns:
        List[Namespaces]: The subworkflows which can be inlined into their parent workflows.
    """
    (step_id, yaml_tree) = yaml_tree_tuple
    yaml_name = step_id.stem

    # Check for top-level yml dsl args
    wic = {'wic': yaml_tree.get('wic', {})}

    if 'implementations' in wic['wic']:
        # Use yaml_name (instead of back_name) and do not append to namespace_init.
        sub_namespaces_list = []
        for stepid, back in wic['wic']['implementations'].items():
            sub_namespaces = get_inlineable_subworkflows(YamlTree(stepid, back), tools, True, namespaces_init)
            sub_namespaces_list.append(sub_namespaces)
        return utils.flatten(sub_namespaces_list)

    steps: List[Yaml] = yaml_tree['steps']
    steps_keys = utils.get_steps_keys(steps)
    subkeys = utils.get_subkeys(steps_keys)

    # All subworkflows are inlineable, except scattered subworkflows.
    inlineable = wic['wic'].get('inlineable', True)
    namespaces = [namespaces_init] if inlineable and namespaces_init != [] and not implementation else []

    for i, step_key in enumerate(steps_keys):
        yaml_stem = Path(yaml_name).stem
        step_name_i = utils.step_name_str(yaml_stem, i, step_key)
        if step_key in subkeys:
            sub_yml_tree = steps[i]['subtree']

            y_t = YamlTree(StepId(step_key, step_id.plugin_ns), sub_yml_tree)
            sub_namespaces = get_inlineable_subworkflows(y_t, tools, False, namespaces_init + [step_name_i])
            namespaces += sub_namespaces

    return namespaces


def inline_subworkflow(yaml_tree_tuple: YamlTree, namespaces: Namespaces) -> Tuple[YamlTree, int]:
    """Inlines the given subworkflow into its immediate parent workflow.

    Args:
        yaml_tree_tuple (YamlTree): A tuple of name and yml AST
        namespaces (Namespaces): Specifies the path in the yml AST to the subworkflow to be inlined.

    Returns:
        YamlTree: The updated root workflow with the given subworkflow inlined into its immediate parent workflow.
    """
    if namespaces == []:
        return yaml_tree_tuple, 0

    (step_id, yaml_tree) = copy.deepcopy(yaml_tree_tuple)
    yaml_name = step_id.stem

    wic = {'wic': yaml_tree.get('wic', {})}
    if 'implementations' in wic['wic']:
        if len(namespaces) == 1:  # and namespaces[0] == yaml_name ?
            (back_name_, yaml_tree) = utils.extract_implementation(yaml_tree, wic['wic'], Path(''))
            yaml_tree = {'steps': yaml_tree['steps']}  # Remove wic tag
            len_substeps = len(yaml_tree['steps'])
            return YamlTree(StepId(back_name_, step_id.plugin_ns), yaml_tree), 0  # len_substeps  # TODO: check step_id

        # Pass namespaces through unmodified
        implementations_trees = []
        for stepid, back in wic['wic']['implementations'].items():
            implementation_tree, len_substeps = inline_subworkflow(YamlTree(stepid, back), namespaces)
            implementations_trees.append(implementation_tree)
        yaml_tree['wic']['implementations'] = dict(implementations_trees)
        return YamlTree(step_id, yaml_tree), 0  # choose len_substeps from which implementation?

    steps: List[Yaml] = yaml_tree['steps']
    steps_keys = utils.get_steps_keys(steps)
    yaml_stem = Path(yaml_name).stem
    step_names = [utils.step_name_str(yaml_stem, i, step_key)
                  for i, step_key in enumerate(steps_keys)]

    if namespaces[0] not in step_names:
        # This should never happen (if namespaces comes from get_inlineable_subworkflows)
        raise Exception(f'Error! {namespaces[0]} not in {step_names}')

    # TODO: We really need to inline the wic tags as well. This may be complicated
    # because due to overloading we may need to modify parent wic tags.

    (yaml_stem, i, step_key) = utils.parse_step_name_str(namespaces[0])
    sub_yml_tree = steps[i]['subtree']
    sub_parentargs = steps[i]['parentargs']

    len_substeps = 0
    if len(namespaces) == 1:
        steps_inits = steps[:i]  # Exclude step i
        steps_tails = steps[i+1:]  # Exclude step i
        # ~ syntax, specifically apply sub_parentargs to all inputs: call sites in sub_yml_tree
        sub_yml_tree = apply_args(sub_yml_tree, sub_parentargs)
        # Inline sub-steps.
        sub_steps: List[Yaml] = sub_yml_tree['steps']
        yaml_tree['steps'] = steps_inits + sub_steps + steps_tails
        # Need to re-index both the sub-step numbers as well as the
        # subsequent steps in this workflow? No, except for wic: steps:
        len_substeps = len(sub_steps)

        parent_wic_tag = wic.get('wic', {}).get("steps", {}).get(
            f'({i + 1}, {step_key})', {}).get('wic', {})
        sub_wic_tag = sub_yml_tree.get('wic', {})

        # TODO: need cleaner code to make arbitrary-depth dictionary.
        if 'wic' not in wic:
            wic['wic'] = {}
        if 'steps' not in wic['wic']:
            wic['wic']['steps'] = {}
        if f'({i + 1}, {step_key})' not in wic['wic']['steps']:
            wic['wic']['steps'][f'({i + 1}, {step_key})'] = {}

        # Merge parent into child to support overloading.
        # TODO: Need to sort the steps by index
        wic['wic']['steps'][f'({i + 1}, {step_key})']['wic'] = \
            merge(sub_wic_tag, parent_wic_tag, strategy=Strategy.TYPESAFE_REPLACE)
    else:
        # Strip off one initial namespace
        y_t = YamlTree(StepId(step_key, step_id.plugin_ns), sub_yml_tree)
        (step_key_, sub_yml_tree), len_substeps = inline_subworkflow(y_t, namespaces[1:])
        # TODO: re-index wic: steps: ? We probably should, although
        # inlineing after merging should not affect CWL args.
        # Re-indexing could be tricky w.r.t. overloading.
        # TODO: maintain inference boundaries (once feature is added)
        # NOTE: Since parentargs are applied after compiling a subworkflow,
        # and since inlineing removes the subworkflow, parentargs does not
        # appear to be inlineing invariant! However, using ~ syntax helps.
        steps[i] = {'id': step_key, 'subtree': sub_yml_tree, 'parentargs': sub_parentargs}

    yaml_tree['wic'] = inline_subworkflow_wic_tag(wic, namespaces, len_substeps)

    return YamlTree(step_id, yaml_tree), len_substeps


def apply_args(sub_yml_tree: Yaml, sub_parentargs: Yaml) -> Yaml:
    # Do we need to deepcopy? We are already deepcopy'ing at the only call site,
    # so looks like no.
    inputs_workflow = sub_yml_tree.get('inputs', {})
    if 'inputs' in sub_yml_tree:
        del sub_yml_tree['inputs']

    steps = sub_yml_tree['steps']
    steps_keys = utils.get_steps_keys(steps)

    for argkey in inputs_workflow:
        # Ordinarily edge inference works across subworkflow boundaries (i.e. is inlineing invariant),
        # but with ~ syntax in the subworkflow and no explicit arguments in the parent workflow,
        # we cannot blindly inline the subworkflow and remove the ~'s in the subworkflow.
        # TODO: Consider adding wic metadata tags to cause inference to skip past the beginning of the subworkflow.
        if argkey not in sub_parentargs.get('in', {}):
            print(f'Warning! Inlineing {argkey} with explicit inputs: in the subworkflow' +
                  'but edge inference in the parent workflow is not supported.')

    for argkey, argval in sub_parentargs.get('in', {}).items():
        # If we are attempting to apply a parameter given in the parent workflow,
        # that parameter had better exist in the subworkflow!
        if not argkey in inputs_workflow:
            raise Exception(f'Error while inlineing {argkey}\n{yaml.dump(sub_yml_tree)}\n{yaml.dump(sub_parentargs)}')

        for i, step_key in enumerate(steps_keys):
            # NOTE: We should probably be using
            # sub_keys = utils.get_subkeys(steps_keys, tools)
            # to check whether or not `step_key in sub_keys` and thus
            # whether or not to use ['parentargs']
            in_step = steps[i].get('in', {})  # CommandLineTools should have ['in'] (if any)
            if not in_step:
                # Subworkflows should have ['parentargs']['in'] (if any)
                in_step = steps[i].get('parentargs', {}).get('in', {})

            for inputkey, inputval in in_step.items():
                if inputval == '~' + argkey:
                    # overwrite ~ syntax / apply argval
                    in_step[inputkey] = argval

    return sub_yml_tree


def inline_subworkflow_wic_tag(wic_tag: Yaml, namespaces: Namespaces, len_substeps: int) -> Yaml:
    """Inlines the wic metadata tags associated with the given subworkflow into its immediate parent wic.

    Args:
        wic_tag (Yaml): The wicmetadata tag associated with the given workflow
        namespaces (Namespaces): Specifies the path in the yml AST to the subworkflow to be inlined.
        len_substeps (int): The number of steps in the subworkflow to be inlined.

    Returns:
        Yaml: The updated wic metadata tag with the wic metadata tag associated with the given subworkflow inlined.
    """
    tag_wic: Yaml = wic_tag['wic']

    # Note: the index after parsing is 0-based.
    step_ints_names = [utils.parse_step_name_str(ns)[1:] for ns in namespaces]

    sub_wic_parent = wic_tag  # initialize to the 'root' wic tag
    # Traverse down to the parent node of the subworkflow to the inlined
    for index, step_name in step_ints_names[:-1]:
        sub_wic_parent = sub_wic_parent.get('wic', {}).get('steps', {}).get(f'({index + 1}, {step_name})', {})
        # Note: if any of the intermediate workflows in the path in the AST tree
        # from the current workflow to the subworkflow being inlined is absent in the current
        # wic metadata tag, the inlining won't have any effect on the wic tag of this workflow.
        # Note: When there're other options like 'graphviz' but not 'steps', we can also short
        # circuit and return.
        if 'steps' not in sub_wic_parent.get('wic', {}):
            return tag_wic  # If path does not exist, do nothing and short circuit

    # Then get the wic tag of the subworkflow
    # Note: sub_index is 0-based.
    sub_index, sub_step_name = step_ints_names[-1]
    sub_wic = sub_wic_parent.get('wic', {}).get('steps', {}).get(f'({sub_index + 1}, {sub_step_name})', {})

    # Note: we should not short circuit when the subworkflow being inlined is not used in the
    # current wic tag, since inlining it will affect the indices of sibling steps following it.
    sub_wic_steps_reindexed = utils.reindex_wic_steps(sub_wic.get('wic', {}).get('steps', {}), 1, sub_index)

    # Delete the subworkflow from the parent workflow since it is replaced by its internal steps.
    # This needs to be explicitly done since the key of this subworkflow in the dict is not
    # the same as any of its inlined steps and therefore won't be overwritten by the deep merge.
    if f'({sub_index + 1}, {sub_step_name})' in sub_wic_parent.get('wic', {}).get('steps', {}):
        del sub_wic_parent['wic']['steps'][f'({sub_index + 1}, {sub_step_name})']

    # The inlining is actually a replacement of the target subworkflows by its steps.
    # Therefore, the incrementing count should be len_substeps - 1.
    sub_wic_parent_steps_reindexed = utils.reindex_wic_steps(sub_wic_parent['wic']['steps'],
                                                             sub_index + 1, len_substeps - 1)

    # Merge the wic: steps: dicts and mutably update the parent
    # Merge parent into child to support overloading.
    # TODO: The 'ranksame' in the wic tag of the inlined subworkflow is ignored
    # and not merged for now.
    sub_wic_parent['wic']['steps'] = merge(sub_wic_steps_reindexed, sub_wic_parent_steps_reindexed,
                                           strategy=Strategy.TYPESAFE_REPLACE)

    return tag_wic


def move_slash_last(source_new: str) -> str:
    """Move / to the last ___ position\n
       (Moving to the last position works because we are inlineing recursively.)

    Args:
        source_new (str): A string representing a CWL dependency, i.e. containing /

    Returns:
        str: source_new with / moved to the last ___ position
    """
    if '/' in source_new:
        source_new = source_new.replace('/', '___')
        source_split = source_new.split('___')
        source_new = '___'.join(source_split[:-1]) + '/' + source_split[-1]
        return source_new

    return source_new


def inline_subworkflow_cwl(rose_tree: RoseTree) -> RoseTree:
    """Inlines all compiled CWL subworkflows into the root workflow.

    Args:
        rose_tree (RoseTree): The data associated with compiled subworkflows

    Returns:
        RoseTree: The updated root workflow with all compiled CWL subworkflows recursively inlined.
    """
    # NOTE: This code is a little bit nasty, and I absolutely do not guarantee that it won't break in the future.
    if rose_tree.sub_trees == []:
        return rose_tree

    sub_trees = [inline_subworkflow_cwl(t) for t in rose_tree.sub_trees]

    node_data: NodeData = rose_tree.data
    cwl_tree = copy.deepcopy(node_data.compiled_cwl)
    # print('cwl_tree', yaml.dump(cwl_tree))
    # print('subtrees')
    # for t in rose_tree.sub_trees:
    #    print(yaml.dump(t.data.compiled_cwl))

    steps = cwl_tree['steps']
    steps_keys = list(steps.keys())
    # NOTE: Only use the last namespace since we are recursively inlineing.
    subkeysdict = {t.data.namespaces[-1]: copy.deepcopy(t.data.compiled_cwl)
                   for t in sub_trees}  # NOT rose_tree.sub_trees
    # print('subkeys', list(subkeysdict.keys()))
    # print('steps_keys', steps_keys)
    steps_new = {}

    count = 0
    for i, step_key in enumerate(steps_keys):
        if step_key in list(subkeysdict.keys()):
            count += 1  # Check that we inline all subworkflows
            inputs = steps[i]['in']
            scattervars = steps[i].get('scatter', [])

            sub_cwl_tree = subkeysdict[step_key]
            sub_steps = sub_cwl_tree['steps']
            sub_steps_new = {}
            for substepkey, substepval in sub_steps.items():
                substep_inputs = substepval['in']
                substep_inputs_new = {}
                for subinputkey, subinputval in substep_inputs.items():
                    # By default, copy the inputs and prepend namespace
                    if isinstance(subinputval, str):
                        source = move_slash_last(subinputval)
                        substep_inputs_new[subinputkey] = step_key + '___' + subinputval

                    if isinstance(subinputval, Dict):
                        source = subinputval['source']
                        source_new = move_slash_last(subinputval['source'])
                        subinputval['source'] = step_key + '___' + source_new
                        substep_inputs_new[subinputkey] = subinputval

                    if source in inputs:
                        # Replace the formal parameter in the subworkflow with
                        # the actual parameter in the parent workflow.
                        newval = inputs[source]

                        if isinstance(newval, str):
                            source_new = move_slash_last(newval)
                            # NOTE: Do not namespace; already namespaced in parent workflow.
                            newval = source_new  # step_key + '___' + source_new

                        if isinstance(newval, Dict) and 'source' in newval:
                            source_new = move_slash_last(newval['source'])
                            # NOTE: Do not namespace; already namespaced in parent workflow.
                            newval['source'] = source_new  # step_key + '___' + source_new

                        substep_inputs_new[subinputkey] = newval  # Overwrite
                        # Copy any input variables referenced, i.e.
                        # initial scatter and/or slice for step 1
                        m = re.match(r'.*\[inputs\.(.*)\].*', str(newval))
                        if m:
                            inputvarname = m.groups()[0]
                            if inputvarname:
                                substep_inputs_new[inputvarname] = inputs[inputvarname]
                                if inputvarname in scattervars:
                                    if 'scatter' in substepval:
                                        substepval['scatter'] += [inputvarname]
                                    else:
                                        substepval['scatter'] = [inputvarname]

                    # Distribute scatter unconditionally across ALL subworkflow dependencies
                    # i.e. https://en.wikipedia.org/wiki/Distributive_property
# NOTE: This code assumes the user has manually performed https://en.wikipedia.org/wiki/Loop-invariant_code_motion
# on the yml. In other words, it assumes that the user has separated / extracted all non-scattered steps from all
# steps that should be scattered. i.e. 1 receptor vs N ligands. Otherwise, we need to transitively follow the edges
# until we can determine the cardinality. It may be possible to avoid the transitive search by bootstrapping in-order,
# but for now let's require the user to manually modify their yml.
                    if scattervars:
                        if ((isinstance(subinputval, str) and '/' in subinputval) or
                                (isinstance(subinputval, Dict) and '/' in subinputval['source'])):
                            if 'scatter' in substepval:
                                if subinputkey not in substepval['scatter']:
                                    substepval['scatter'] += [subinputkey]
                            else:
                                substepval['scatter'] = [subinputkey]
                            substepval['scatterMethod'] = 'dotproduct'

                # Overwrite inputs
                substepval['in'] = substep_inputs_new

                # Modify run tag
                runstr = substepval['run']
                if runstr.startswith('../'):
                    substepval['run'] = runstr[len('../'):]
                # TODO: Consider general case of prepending namespace / directory

                # prepend namespace to step names
                namespaced = step_key + '___' + substepkey
                sub_steps_new[namespaced] = substepval

            sub_cwl_tree['steps'] = sub_steps_new

            # Insert the steps from the subworkflow
            steps_new.update(sub_steps_new)
        else:
            # Otherwise, just copy the step
            steps_new[i] = steps[i]

    if count != len(subkeysdict):
        print('Error! Not all subworkflows inlined!')

    cwl_tree['steps'] = steps_new

    # Finally, for all outputs in the parent workflow
    outputs = cwl_tree['outputs']
    outputs_new = {}
    for outkey, outval in outputs.items():
        if 'output_all' in outkey:
            continue  # Skip for now.

        outval['outputSource'] = move_slash_last(outval['outputSource'])
        outputs_new[outkey] = outval

    cwl_tree['outputs'] = outputs_new

    data = NodeData(node_data.namespaces, node_data.name, node_data.yml, cwl_tree,  # NOTE: Only updating cwl_tree
                    node_data.tool, node_data.workflow_inputs_file, node_data.explicit_edge_defs,
                    node_data.explicit_edge_calls, node_data.graph,
                    node_data.inputs_workflow, node_data.step_name_1)

    # print('cwl_tree', yaml.dump(cwl_tree))
    # print('subtrees')
    # for t in rose_tree.sub_trees:
    #    print(yaml.dump(t.data.compiled_cwl))

    return RoseTree(data, [])



import argparse
import sys
from pathlib import Path
from unittest.mock import patch

from . import _version
__version__ = _version.get_versions()['version']

parser = argparse.ArgumentParser(prog='main', description='Convert a high-level yaml workflow file to CWL.')
parser.add_argument('--yaml', type=str, required=('--generate_schemas' not in sys.argv),
                    help='Yaml workflow file')
parser.add_argument('--inputs_file', type=str, required=False, default='',
                    help='Additional inputs Yaml file')
parser.add_argument('--config_file', type=str, required=False, default=str(Path().home()/'wic'/'global_config.json'),
                    help='User provided (JSON) config file')
# version action exits the parser Ref : https://github.com/python/cpython/blob/1f515e8a109204f7399d85b7fd806135166422d9/Lib/argparse.py#L1167
parser.add_argument('--version', action='version', version=__version__,
                    default='==SUPPRESS==', help='Current version of the Workflow Inference Compiler')
parser.add_argument('--generate_schemas', default=False, action="store_true",
                    help='Generate schemas for the files in config.json (search_paths_wic and search_paths_cwl)')
parser.add_argument('--homedir', type=str, required=False, default=str(Path().home()),
                    help='The users home directory. This is necessary because CWL clears environment variables (e.g. HOME)')
parser.add_argument('--insert_steps_automatically', default=False, action="store_true",
                    help='''Attempt to fix inference failures by speculatively
                    inserting workflow steps from a curated whitelist.''')
parser.add_argument('--no_provenance', default=False, action="store_true",
                    help='Do not use the --provenance feature of CWL. (You should only disable provenance if absolutely necessary.)')
parser.add_argument('--copy_output_files', default=False, action="store_true",
                    help='Copies output files from the cachedir to outdir/ (automatically enabled with --run_local)')
parser.add_argument('--inline_cwl_runtag', default=False, action="store_true",
                    help='Copies cwl adapter file contents inline into the final .cwl in autogenerated/')

parser.add_argument('--parallel', default=False, action="store_true",
                    help='''When running locally, execute independent steps in parallel.
                    \nThis is required for real-time analysis, but it may cause issues with
                    \nhanging (particularly when scattering). See user guide for details.''')
parser.add_argument('--quiet', default=False, action="store_true",
                    help='''Disable verbose output. This will not print out the commands used for each step,
                    and it will capture all stdout/stderr into log files for each step.''')
parser.add_argument('--cwl_runner', type=str, required=False, default='cwltool', choices=['cwltool', 'toil-cwl-runner'],
                    help='The CWL runner to use for running workflows locally.')
parser.add_argument('--allow_raw_cwl', default=False, action="store_true",
                    help='Do not check whether the input to a workflow step refers to the workflow inputs: tag')
parser.add_argument('--ignore_docker_install', default=False, action="store_true",
                    help='''Do not check whether docker is installed before running workflows.
                    \n--ignore_docker_install does NOT change whether or not any step in your workflow uses docker!''')
parser.add_argument('--ignore_docker_processes', default=False, action="store_true",
                    help='Do not check whether there are too many running docker processes before running workflows.')
parser.add_argument('--container_engine', default='docker',
                    help='Specify which command to use to run OCI containers.')

parser.add_argument('--docker_remove_entrypoints', default=False, action="store_true",
                    help='''Remove entrypoints from docker images before running workflows.
                    See https://cwl.discourse.group/t/override-docker-entrypoint-in-command-line-tool/695/2
                    If your CWL CommandLineTools rely on non-portable docker entrypoints
                    (i.e. if they do NOT explicitly have baseCommand: and/or arguments: tags)
                    then this flag will almost certainly cause your workflows to fail.
                    In other words, this flag can be used to tell you if your workflows are
                    trapped inside a walled garden of docker containers or if there is at least a chance that
                    they can be run in dev environments (conda, etc) and/or directly on the host machine.''')

parser.add_argument('--partial_failure_enable', default=False, action="store_true",
                    help='''Let workflows to continue with partial failures
                    \n(i.e. run next steps after failure step by guarding for possible no input)''')

parser.add_argument('--partial_failure_success_codes', nargs='*', type=int, default=[0, 1], required=False,
                    help='Let users add custom error codes to be treated as success')

parser.add_argument('--partial_failure_success_codes_range', nargs=2, type=int, default=[0, 1], required=False,
                    help='Let users add custom error code range to be treated as success')

group_run = parser.add_mutually_exclusive_group()
group_run.add_argument('--generate_run_script', default=False, action="store_true",
                       help='Just generates run.sh and exits. Does not actually invoke ./run.sh')
group_run.add_argument('--run_local', default=False, action="store_true",
                       help='After generating the cwl file(s), run it on your local machine.')

parser.add_argument('--cwl_inline_subworkflows', default=False, action="store_true",
                    help='Before generating the cwl file, inline all subworkflows.')
parser.add_argument('--inference_disable', default=False, action="store_true",
                    help='Disables use of the inference algorithm when compiling.')
parser.add_argument('--inference_use_naming_conventions', default=False, action="store_true",
                    help='Enables the use of naming conventions in the inference algorithm')
parser.add_argument('--validate_plugins', default=False, action="store_true",
                    help='Validate all CWL CommandLineTools')
parser.add_argument('--ignore_validation_errors', default=False, action="store_true",
                    help='Temporarily ignore validation errors. Do not use this permanently!')
parser.add_argument('--no_skip_dollar_schemas', default=False, action="store_true",
                    help='''Does not skip processing $schemas tags in CWL files for performance.
                    Skipping significantly improves initial validation performance, but is not always desired.
                    See https://github.com/common-workflow-language/cwltool/issues/623''')
parser.add_argument('--cachedir', type=str, required=False, default='cachedir',
                    help='The directory to save intermediate results; useful with RealtimePlots.py')

parser.add_argument('--graphviz', default=False, action="store_true",
                    help='Generate a DAG using graphviz.')
parser.add_argument('--graph_label_edges', default=False, action="store_true",
                    help='Label the graph edges with the name of the intermediate input/output.')
parser.add_argument('--graph_label_stepname', default=False, action="store_true",
                    help='Prepend the step name to each step node.')
parser.add_argument('--graph_show_inputs', default=False, action="store_true",
                    help='Add nodes to the graph representing the workflow inputs.')
parser.add_argument('--graph_show_outputs', default=False, action="store_true",
                    help='Add nodes to the graph representing the workflow outputs.')
parser.add_argument('--graph_inline_depth', type=int, required=False, default=sys.maxsize,
                    help='Controls the depth of subgraphs which are displayed.')
parser.add_argument('--graph_dark_theme', default=False, action="store_true",
                    help='Changees the color of the fonts and edges from white to black.')
parser.add_argument('--custom_net', type=str, required=False,
                    help='Passes --custom-net flag to cwltool.')


def get_args(yaml_path: str = '', suppliedargs: list[str] = []) -> argparse.Namespace:
    """This is used to get mock command line arguments, default + suppled args

    Returns:
        argparse.Namespace: The mocked command line arguments
    """
    defaultargs = ['wic', '--yaml', yaml_path]  # ignore --yaml
    testargs = defaultargs + suppliedargs
    with patch.object(sys, 'argv', testargs):
        args: argparse.Namespace = parser.parse_args()
    return args



import argparse
import copy
from shutil import copytree, ignore_patterns
import json
from pathlib import Path
import sys
from typing import Any, List, Tuple

import yaml

from . import auto_gen_header
from .wic_types import (Namespaces, NodeData, RoseTree, Yaml, ExplicitEdgeCalls, Json)


def read_lines_pairs(filename: Path) -> List[Tuple[str, str]]:
    """Reads a whitespace-delimited file containing two paired entries per line (i.e. a serialized Dict).

    Args:
        filename (Path): The full path of the file to be read.

    Raises:
        Exception: If any non-blank, non-comment lines do not contain exactly two entries.

    Returns:
        List[Tuple[str, str]]: The file contents, with blank lines and comments removed.
    """
    with open(filename, mode='r', encoding='utf-8') as f:
        lines = []
        for line in f.readlines():
            if line.strip() == '':  # Skip blank lines
                continue
            if line.startswith('#'):  # Skip comment lines
                continue
            l_s = line.split()
            if not len(l_s) == 2:
                print(line)
                raise Exception("Error! Line must contain exactly two entries!")
            lines.append((l_s[0], l_s[1]))
    return lines


# snakeyaml (a cromwell dependency) refuses to parse yaml files with more than
# 50 anchors/aliases to prevent Billion Laughs attacks.
# See https://en.wikipedia.org/wiki/Billion_laughs_attack
# Solution: Inline the contents of the aliases into the anchors.
# See https://ttl255.com/yaml-anchors-and-aliases-and-how-to-disable-them/#override


class NoAliasDumper(yaml.SafeDumper):
    def ignore_aliases(self, data: Any) -> bool:
        return True


def write_to_disk(rose_tree: RoseTree, path: Path, relative_run_path: bool, inputs_file: str = '') -> None:
    """Writes the compiled CWL files and their associated yml inputs files to disk.

    NOTE: Only the yml input file associated with the root workflow is
    guaranteed to have all inputs. In other words, subworkflows will all have
    valid CWL files, but may not be executable due to 'missing' inputs.
    Additional inputs can be explicitly passed in in this case.

    Args:
        rose_tree (RoseTree): The data associated with compiled subworkflows
        path (Path): The directory in which to write the files
        relative_run_path (bool): Controls whether to use subdirectories or just one directory.
        inputs_file (str): Optional additional inputs
    """
    inputs = {}
    if inputs_file:
        with open(inputs_file, mode='r', encoding='utf-8') as f:
            inputs = yaml.safe_load(f.read())
        for key, val in inputs.items():
            if 'location' in val and not Path(val['location']).is_absolute():
                # Change relative paths for class: File and class: Dir
                # to be w.r.t. autogenerated/
                val['location'] = '../' + val['location']
    _write_to_disk(rose_tree, path, relative_run_path, inputs)


def _write_to_disk(rose_tree: RoseTree, path: Path, relative_run_path: bool, inputs: Yaml = {}) -> None:
    """Writes the compiled CWL files and their associated yml inputs files to disk.

    NOTE: Only the yml input file associated with the root workflow is
    guaranteed to have all inputs. In other words, subworkflows will all have
    valid CWL files, but may not be executable due to 'missing' inputs.
    Additional inputs can be explicitly passed in in this case.

    Args:
        rose_tree (RoseTree): The data associated with compiled subworkflows
        path (Path): The directory in which to write the files
        relative_run_path (bool): Controls whether to use subdirectories or just one directory.
        inputs (Yaml): Optional additional inputs
    """
    node_data: NodeData = rose_tree.data
    namespaces = node_data.namespaces
    yaml_stem = node_data.name
    cwl_tree = node_data.compiled_cwl
    yaml_inputs = {**node_data.workflow_inputs_file, **inputs}

    path.mkdir(parents=True, exist_ok=True)
    if relative_run_path:
        filename_cwl = f'{yaml_stem}.cwl'
        filename_yml = f'{yaml_stem}_inputs.yml'
    else:
        filename_cwl = '___'.join(namespaces + [f'{yaml_stem}.cwl'])
        filename_yml = '___'.join(namespaces + [f'{yaml_stem}_inputs.yml'])

    # Dump the compiled CWL file contents to disk.
    # Use sort_keys=False to preserve the order of the steps.
    yaml_content = yaml.dump(cwl_tree, sort_keys=False, line_break='\n', indent=2, Dumper=NoAliasDumper)
    with open(path / filename_cwl, mode='w', encoding='utf-8') as w:
        w.write('#!/usr/bin/env cwl-runner\n')
        w.write(auto_gen_header)
        w.write(''.join(yaml_content))

    yaml_content = yaml.dump(yaml_inputs, sort_keys=False, line_break='\n', indent=2, Dumper=NoAliasDumper)
    with open(path / filename_yml, mode='w', encoding='utf-8') as inp:
        inp.write(auto_gen_header)
        inp.write(yaml_content)

    for sub_rose_tree in rose_tree.sub_trees:
        subpath = path
        if relative_run_path:
            sub_node_data: NodeData = sub_rose_tree.data
            sub_step_name = sub_node_data.namespaces[-1]
            subpath = path / sub_step_name
        _write_to_disk(sub_rose_tree, subpath, relative_run_path, inputs)


def write_config_to_disk(config: Json, config_file: Path) -> None:
    """Writes config json object to config_file

    Args:
        config (Json): The json object that is to be written to disk
        config_file (Path): The file path where it is to be written
    """
    config_dir = Path(config_file).parent
    # make the full path if it doesn't exist
    config_dir.mkdir(parents=True, exist_ok=True)
    with open(config_file, 'w', encoding='utf-8') as f:
        json.dump(config, f)


def get_config(config_file: Path, default_config_file: Path) -> Json:
    """Returns the config json object from config_file with absolute paths

    Args:
        config_file (Path): The path of the user specified config file
        default_config_file (Path): The default path of the config file if user hasn't specified one

    Returns:
        Json: The config json object with absolute filepaths
    """
    global_config: Json = {}
    proj_root_dir = 'workflow-inference-compiler'
    if not config_file.exists():
        # check if sophios is run from  'project root' dir
        if str(Path.cwd()).endswith(proj_root_dir):
            if config_file == default_config_file:
                global_config = get_default_config()
                # write the default config object to the 'global_config.json' file in user's ~/wic directory
                # for user to inspect and or modify the config json file
                write_config_to_disk(global_config, default_config_file)
                print(f'default config file : {default_config_file} generated')
            else:
                print(f"Error user specified config file {config_file} doesn't exist")
                sys.exit()
        else:
            global_config = get_basic_config()
            # write the basic config object to the 'global_config.json' file in user's ~/wic directory
            # for user to inspect and or modify the config json file
            write_config_to_disk(global_config, default_config_file)
            move_adapters_and_examples(global_config)
            print(f'default config file : {default_config_file} generated')
    else:
        # reading user specified config file only if it exists
        # never overwrite user's config file or generate another file in user's non-default directory
        # TODO : Validate the json inside 'read_config_from_disk' function
        global_config = read_config_from_disk(config_file)
    return global_config


def move_adapters_and_examples(config: Json) -> None:
    """Copies all the adapters and examples into the given search paths

    Args:
        config (Json): Json containing search path information
    """
    # if the sophios installation is through pip package
    pip_dir_name = 'sophios'
    if str(Path(__file__).parent).endswith(pip_dir_name):
        adapters_dir = Path(__file__).parent / 'cwl_adapters'
        examples_dir = Path(__file__).parent / 'examples'
    # when using dev/git install
    else:
        adapters_dir = Path(__file__).parent.parent.parent / 'cwl_adapters'
        examples_dir = Path(__file__).parent.parent.parent / 'docs' / 'tutorials'

    extlist = ['*.png', '*.md', '*.rst', '*.pyc', '__pycache__', '*.json']
    # the default search paths will be the first entry in the 'global' namespace
    # but we also want to preserve the ability of the users to add more search paths
    # so we keep the namespace tag as a list of paths rather than a single path
    copytree(adapters_dir, Path(config['search_paths_cwl']['global'][0]))
    copytree(examples_dir, Path(config['search_paths_wic']['global'][0]),
             ignore=ignore_patterns(*extlist))


def read_config_from_disk(config_file: Path, abspath: bool = True) -> Json:
    """Returns the config json object from config_file with absolute paths

    Args:
        config_file (Path): The path of json file where it is to be read from

    Returns:
        Json: The config json object with absolute filepaths
    """
    config: Json = {}
    # config_file can contain absolute or relative paths
    with open(config_file, 'r', encoding='utf-8') as f:
        config = json.load(f)
    conf_tags = ['search_paths_cwl', 'search_paths_wic']
    for tag in conf_tags:
        if abspath:
            config[tag] = get_absolute_paths(config[tag])
        else:  # this is a hacky way to fix global paths wrt ~/home/wic/
            config[tag] = get_home_paths(config[tag])
    return config


def get_default_config() -> Json:
    """Returns the default config with absolute paths

    Returns:
        Json: The config json object with absolute filepaths
    """
    src_dir = Path(__file__).parent
    default_config: Json = {}
    # read_config_from_disk handles converting them to absolute paths
    default_config = read_config_from_disk(src_dir/'config.json')
    return default_config


def get_basic_config() -> Json:
    """Returns the (default) basic config with absolute paths

    Returns:
        Json: The config json object with absolute filepaths
    """
    # basic config doesn't contain mm-workflows or other
    # domain cwl or wic files search paths
    src_dir = Path(__file__).parent
    basic_config: Json = {}
    # read_config_from_disk handles converting them to absolute paths
    basic_config = read_config_from_disk(src_dir/'config_basic.json', False)
    return basic_config


def get_absolute_paths(sub_config: Json) -> Json:
    """Update the paths within the sub_config json object as absolute paths

    Args:
        sub_config (dict): The json (sub)object where filepaths are stored

    Returns:
        Json: The json (sub)object with absolute filepaths
    """
    abs_sub_config = copy.deepcopy(sub_config)
    for ns in abs_sub_config:
        abs_paths = [str(Path(path).absolute()) for path in abs_sub_config[ns]]
        abs_sub_config[ns] = abs_paths
    return abs_sub_config


def get_home_paths(sub_config: Json) -> Json:
    """Update the paths within the sub_config json object as absolute paths

    Args:
        sub_config (dict): The json (sub)object where filepaths are stored

    Returns:
        Json: The json (sub)object with absolute filepaths
    """
    abs_sub_config = copy.deepcopy(sub_config)
    for ns in abs_sub_config:
        abs_paths = [str(Path.home() / path) for path in abs_sub_config[ns]]
        abs_sub_config[ns] = abs_paths
    return abs_sub_config


def write_absolute_yaml_tags(args: argparse.Namespace, in_dict_in: Yaml, namespaces: Namespaces,
                             step_name_i: str, explicit_edge_calls_copy: ExplicitEdgeCalls) -> None:
    """cwl_subinterpreter requires all paths to be absolute.

    Args:
        args (argparse.Namespace): The command line arguments
        in_dict_in (Yaml): The in: subtag of a cwl_subinterpreter: tag. (Mutates in_dict_in)
        namespaces (Namespaces): Specifies the path in the yml AST to the current subworkflow
        step_name_i (str): The name of the current workflow step
        explicit_edge_calls_copy (ExplicitEdgeCalls): Stores the (path, value) of the explicit edge call sites
    """

    # cachedir_path needs to be an absolute path, but for reproducibility
    # we don't want users' home directories in the yml files.
    cachedir_path = Path(args.cachedir).absolute()
    # print('setting cachedir_path to', cachedir_path)
    in_dict_in['root_workflow_yml_path'] = {'wic_inline_input': str(Path(args.yaml).parent.absolute())}

    in_dict_in['cachedir_path'] = {'wic_inline_input': str(cachedir_path)}
    in_dict_in['homedir'] = {'wic_inline_input': args.homedir}

    # Add a 'dummy' values to explicit_edge_calls, because
    # that determines sub_args_provided when the recursion returns.
    arg_keys_ = ['root_workflow_yml_path', 'cachedir_path', 'homedir']
    for arg_key_ in arg_keys_:
        in_name_ = f'{step_name_i}___{arg_key_}'  # {step_name_i}_input___{arg_key}
        explicit_edge_calls_copy.update({in_name_: (namespaces + [step_name_i], arg_key_)})



import argparse
import glob
import json
import os
import subprocess as sub
import sys
import time
from pathlib import Path
from typing import Dict, List
from unittest.mock import patch

import graphviz
import networkx as nx
from jsonschema import Draft202012Validator

from . import input_output as io
from . import ast, cli, compiler, inference, utils
from .run_local import stage_input_files
from .plugins import get_tools_cwl, get_yml_paths, logging_filters
from .schemas import wic_schema
from .wic_types import GraphData, GraphReps, Json, StepId, Tools, YamlTree

# from watchdog.observers import Observer
# from watchdog.observers.polling import PollingObserver
# from watchdog.events import FileSystemEvent, PatternMatchingEventHandler


def absolute_paths(config: Json, cachedir_path: Path) -> Json:
    """Recursively searches for paths in config and makes them absolute by prepending cachedir_path.

    Args:
        config (Json): The contents of the YAML cwl_subinterpreter config: tag.
        cachedir_path (Path): The --cachedir directory of the main workflow.

    Returns:
        Json: The contents of the YAML cwl_subinterpreter config: tag, with all paths prepended with cachedir_path.
    """
    new_json: Json = {}
    for key, val in config.items():
        if isinstance(val, Dict):
            new_val = absolute_paths(val, cachedir_path)
        else:
            new_val = val
            # TODO: Improve this heuristic
            if 'input' in key and 'path' in key:
                new_val = str(cachedir_path / val)  # type: ignore
                changed_files = file_watcher_glob(cachedir_path, val, {})
                # We require unique filenames, so there should only be one file.
                # (except for files that get created within the cwl_subinterpreter workflow itself)
                changed_files_lst = list(changed_files.items())
                if len(changed_files_lst) == 0:
                    print(f'Warning! Changed files should be length one! {val}\n{changed_files_lst}')
                else:
                    if len(changed_files_lst) != 1:
                        print(f'Warning! Changed files should be length one! {val}\n{changed_files_lst}')
                    changed_files_lst.sort(key=lambda x: x[1])
                    file = changed_files_lst[-1][0]  # most recent
                    new_val = str(Path(file).absolute())  # type: ignore
        new_json[key] = new_val
    return new_json


def rerun_cwltool(homedir: str, _directory_realtime: Path, cachedir_path: Path, cwl_tool: str,
                  args_vals: Json, tools_cwl: Tools, yml_paths: Dict[str, Dict[str, Path]],
                  validator: Draft202012Validator, root_workflow_yml_path: Path) -> None:
    """This will speculatively execute cwltool for real-time analysis purposes.\n
    It will NOT check for return code 0. See docs/userguide.md

    Args:
        homedir (str): The users home directory
        _directory_realtime (Path): The working directory of the main workflow.\n
        Currently unused to avoid this workflow from overwriting files from the main\n
        workflow (which by design will likely be running concurrently with this code).
        cachedir_path (Path): The --cachedir directory of the main workflow.
        cwl_tool (str): The CWL CommandLineTool or YAML filename (without extension).
        args_vals (Json): The contents of the YAML cwl_subinterpreter config: tag.
        tools_cwl (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()
        yml_paths (Dict[str, Dict[str, Path]]): The yml workflow definitions found using get_yml_paths()
        validator (Draft202012Validator): Used to validate the yml files against the autogenerated schema.
        root_workflow_yml_path (Path): The full absolute path to the root workflow yml file.
    """
    try:
        # Make paths in arguments absolute w.r.t the realtime directory. See below.
        args_vals_new = absolute_paths(args_vals, cachedir_path)

        # Construct a single-step workflow and add its arguments
        # import yaml
        if Path(cwl_tool).suffix == '.wic':
            yaml_path = cwl_tool
            wic_steps = {'steps': {f'(1, {cwl_tool})': {'wic': {'steps': args_vals_new}}}}
            root_yaml_tree = {'wic': wic_steps, 'steps': [{cwl_tool: None}]}
            # print('root_yaml_tree')
            # print(yaml.dump(root_yaml_tree))
            # TODO: Support other namespaces
            plugin_ns = 'global'  # wic['wic'].get('namespace', 'global')
            step_id = StepId(yaml_path, plugin_ns)
            y_t = YamlTree(step_id, root_yaml_tree)
            yaml_tree_raw = ast.read_ast_from_disk(homedir, y_t, yml_paths, tools_cwl, validator, True)
            yaml_tree = ast.merge_yml_trees(yaml_tree_raw, {}, tools_cwl)
            yaml_tree = ast.python_script_generate_cwl(yaml_tree, Path(''), tools_cwl)
            yml = yaml_tree.yml
        else:
            yml = {'steps': [{cwl_tool: args_vals_new}]}
        # print('yml')
        # print(yml)
        # print(yaml.dump(yml))

        # Measure compile time
        time_initial = time.time()

        # Setup dummy args
        args = cli.get_args()

        # TODO: Support other namespaces
        plugin_ns = 'global'  # wic['wic'].get('namespace', 'global')
        yaml_path = f'{cwl_tool}_only.wic'
        stepid = StepId(yaml_path, plugin_ns)
        yaml_tree = YamlTree(stepid, yml)
        subgraph = GraphReps(graphviz.Digraph(name=yaml_path), nx.DiGraph(), GraphData(yaml_path))
        compiler_info = compiler.compile_workflow(yaml_tree, args, [], [subgraph], {}, {}, {}, {},
                                                  tools_cwl, True, relative_run_path=False, testing=False)
        rose_tree = compiler_info.rose
        working_dir = Path('.') / Path('autogenerated/')  # Use a new working directory.
        # Can also use `_directory_realtime` / Path('autogenerated/') at the risk of overwriting other files.
        io.write_to_disk(rose_tree, working_dir, False, args.inputs_file)

        time_final = time.time()
        print(f'compile time for {cwl_tool}: {round(time_final - time_initial, 4)} seconds')

        yaml_inputs = rose_tree.data.workflow_inputs_file
        stage_input_files(yaml_inputs, root_workflow_yml_path, relative_run_path=False, throw=False)

        # NOTE: Since we are running cwltool 'within' cwltool, the inner
        # cwltool command will get run from working_dir, but then cwl_tool
        # will run within some other hashed directory in .../cachedir/
        # The solution is to modify the input paths above to be absolute.
        # The easiest way to do this for now is recompiling. This adds a few
        # seconds, but most of the time will be CWL validation and runtime.
        # Alternatively, we could try to compile once in main() and then
        # make the paths absolute in f'{cwl_tool}_only_inputs.yml' here.
        cmd: List[str] = ['cwltool', '--cachedir', str(cachedir_path),
                          f'{cwl_tool}_only.cwl', f'{cwl_tool}_only_inputs.yml']
        # proc = sub.run(self.cmd, cwd=working_dir)
        # cmd = self.cmd
        print('Running', cmd)
        proc = sub.run(cmd, cwd=working_dir, check=False)  # See below!
        print('inner cwltool completed')
        # Don't check the return code because the file may not exist yet, or
        # because speculative execution may fail for any number of reasons.
        # proc.check_returncode()
    except FileNotFoundError as e:
        # The file may not exist yet.
        print(e)


# NOTE: You should be very careful when using file watchers! Most libraries
# (watchdog, watchfiles, etc) will use operating system / platform-specific
# APIs to check for changes (for performance reasons). However, this can cause
# problems in some cases, specifically for network drives.
# See https://stackoverflow.com/questions/45441623/using-watchdog-of-python-to-monitoring-afp-shared-folder-from-linux
# I'm 99% sure the same problem happens with Docker containers. Either way, the
# solution is to use polling. However, for unknown reasons, simply replacing
# Observer with PollingObserver doesn't seem to work! So we are forced to write
# our own basic file watcher using glob.


# class SubprocessHandler(PatternMatchingEventHandler):

#     def __init__(self, cmd: List[str], cachedir_path: str, cwl_tool: str, args_vals: Json, tools_cwl: Tools, **kwargs: Any) -> None:
#         self.cmd = cmd
#         self.lock = False
#         self.cachedir_path = cachedir_path
#         self.cwl_tool = cwl_tool
#         self.args_vals = args_vals
#         self.tools_cwl = tools_cwl
#         super().__init__(**kwargs)

#     def on_any_event(self, event: FileSystemEvent) -> None:
#         # Use a lock to prevent us from DOS'ing ourselves
#         global lock
#         if event.event_type == 'modified' and not lock:
#             directory = Path(event._src_path).parent
#             print('directory', directory)
#             # self.lock = True
#             print(event)
#             rerun_cwltool(directory, self.cachedir_path, self.cwl_tool, self.args_vals, self.tools_cwl)
#             # self.lock = False


def file_watcher_glob(cachedir_path: Path, pattern: str, prev_files: Dict[str, float]) -> Dict[str, float]:
    """Determines whether files (specified by the given glob pattern) have been either recently created or modified.\n
    Note that this is a workaround due to an issue with using standard file-watching libraries.

    Args:
        cachedir_path (Path): The --cachedir directory of the main workflow.
        pattern (str): The glob pattern which specifies the files to be watched.
        prev_files (Dict[str, float]): This should be the return value from the previous function call.

    Returns:
        Dict[str, float]: A dictionary containing the filepaths and last modification times.
    """
    changed_files = {}
    file_pattern = str(cachedir_path / f'**/{pattern}')
    file_paths = glob.glob(file_pattern, recursive=True)
    for file in file_paths:
        mtime = os.path.getmtime(file)
        if file not in prev_files:
            # created
            changed_files[file] = mtime
        elif mtime > prev_files[file]:
            # modified
            changed_files[file] = mtime
    return changed_files


def cli_watcher() -> argparse.Namespace:
    """This contains the command line arguments for speculative execution.

    Returns:
        argparse.Namespace: The command line arguments
    """
    parser = argparse.ArgumentParser(prog='main', description='Speculatively execute a high-level yaml workflow file.')
    parser.add_argument('--cwl_tool', type=str, required=True,
                        help='CWL or WIC filestem')
    parser.add_argument('--cachedir_path', type=str, required=True,
                        help='This should be set to the --cachedir option from the main cli')
    parser.add_argument('--file_pattern', type=str, required=True,
                        help='This glob pattern is used to find files within --cachedir_path')
    parser.add_argument('--max_times', type=str, required=True,
                        help='--cwl_tool will be speculatively executed at most max_times')
    parser.add_argument('--config', type=str, required=True,
                        help='This should be a json-encoded representation of the config: YAML subtag of --cwl_tool')
    parser.add_argument('--homedir', type=str, required=False,  # default=str(Path().home()), # NOTE: no default!
                        help='The users home directory. This is necessary because CWL clears environment variables (e.g. HOME)')
    parser.add_argument('--root_workflow_yml_path', type=str, required=True,
                        help='The full absolute path to the root workflow yml file.')
    return parser.parse_args()


def main() -> None:
    """See docs/userguide.md#real-time-analysis--speculative-execution"""
    print('cwl_subinterpreter sys.argv', sys.argv)
    args = cli_watcher()
    logging_filters()

    cachedir_path = Path(args.cachedir_path)
    file_pattern = args.file_pattern
    cwl_tool = args.cwl_tool
    max_times = int(args.max_times)
    root_workflow_yml_path = Path(args.root_workflow_yml_path)

    # Create an empty 'logfile' so that cwl_subinterpreter.cwl succeeds.
    # TODO: Maybe capture cwl_tool stdout/stderr and redirect to this logfile.
    logfile = Path(f'{cwl_tool}_only.log')
    logfile.touch()

    args_vals = json.loads(args.config)
    # In CWL all env variables are hidden by default so Path().home() doesn't work
    # Also User may specify a different homedir
    default_config_file = Path(args.homedir)/'wic'/'global_config.json'
    global_config: Json = io.get_config(Path(args.config_file), default_config_file)

    tools_cwl = get_tools_cwl(global_config, quiet=args.quiet)
    yml_paths = get_yml_paths(global_config)

    # Perform initialization via mutating global variables (This is not ideal)
    compiler.inference_rules = global_config.get('inference_rules', {})
    inference.renaming_conventions = global_config.get('renaming_conventions', [])

    # Generate schemas for validation
    yaml_stems = utils.flatten([list(p) for p in yml_paths.values()])
    validator = wic_schema.get_validator(tools_cwl, yaml_stems)

    cachedir_hash_path = Path('.').absolute()
    print('cachedir_hash_path', cachedir_hash_path)

    """cmd: List[str] = ['cwltool', '--cachedir', cachedir_path,
                         f'{cachedir_hash_path}/{cwl_tool}_only.cwl',
                         f'{cachedir_hash_path}/{cwl_tool}_only_inputs.yml']
    event_handler = SubprocessHandler(cmd, cachedir_path, cwl_tool, args_vals, tools_cwl, patterns=[file_pattern])
    observer = PollingObserver()  # This does not work!
    observer.schedule(event_handler, cachedir_path, recursive=True)
    observer.start()"""

    # Specify a maximum number of iterations to guarantee termination.
    # Total runtime will be (sleep time + compile time + run time) * max_iters
    # For now, there is no way to estimate max_iters such that polling will end
    # around the same time as the original workflow step.
    # TODO: Generate a file when the original workflow step finishes, and look
    # for that file here to terminate. Keep max_iters just in case.
    i = 0
    prev_files: Dict[str, float] = {}
    try:
        while i < max_times:
            # Use our own polling file watcher, see above.
            changed_files = file_watcher_glob(cachedir_path, file_pattern, prev_files)
            # print('len(changed_files)', len(changed_files))
            for file in changed_files:
                if file_pattern[1:] in file:
                    print(file)
                    rerun_cwltool(args.homedir, Path(file).parent, cachedir_path, cwl_tool,
                                  args_vals, tools_cwl, yml_paths, validator,
                                  root_workflow_yml_path)
            prev_files = {**prev_files, **changed_files}

            time.sleep(1.0)  # Wait at least 1 second so we don't just spin.
            i += 1
    except KeyboardInterrupt:
        pass
    # observer.stop()
    # observer.join()

    failed = False  # Your analysis goes here
    if failed:
        print(f'{cwl_tool} failed!')
        sys.exit(1)


if __name__ == "__main__":
    main()



import json
# import subprocess as sub
from pathlib import Path
# import signal
# import sys
# from typing import List
# import argparse
import asyncio
from jsonschema import Draft202012Validator

from fastapi import Request

import pytest
from sophios.wic_types import Json


from sophios.api.http import restapi

SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "definitions": {
        "Link": {
            "properties": {
                "id": {
                    "type": "number"
                },
                "inletIndex": {
                    "type": "number"
                },
                "outletIndex": {
                    "type": "number"
                },
                "sourceId": {
                    "type": "number"
                },
                "targetId": {
                    "type": "number"
                },
                "x1": {
                    "type": "number"
                },
                "x2": {
                    "type": "number"
                },
                "y1": {
                    "type": "number"
                },
                "y2": {
                    "type": "number"
                }
            },
            "type": "object",
            "required": ["id", "inletIndex", "outletIndex", "sourceId", "targetId"]
        },
        "NodeSettings": {
            "properties": {
                "inputs": {
                    "additionalProperties": {
                        "$ref": "#/definitions/T"
                    },
                    "type": "object"
                },
                "outputs": {
                    "additionalProperties": {
                        "$ref": "#/definitions/T"
                    },
                    "type": "object"
                }
            },
            "type": "object"
        },
        "NodeX": {
            "properties": {
                "expanded": {
                    "type": "boolean"
                },
                "height": {
                    "type": "number"
                },
                "id": {
                    "type": "number"
                },
                "internal": {
                    "type": "boolean"
                },
                "name": {
                    "type": "string"
                },
                "pluginId": {
                    "type": "string"
                },
                "settings": {
                    "$ref": "#/definitions/NodeSettings"
                },
                "width": {
                    "type": "number"
                },
                "x": {
                    "type": "number"
                },
                "y": {
                    "type": "number"
                },
                "z": {
                    "type": "number"
                },
            },
            "type": "object",
            "required": ["id", "name", "pluginId", "settings", "internal"]
        },
        "T": {
            "type": "object"
        }
    },
    "properties": {
        "links": {
            "items": {
                "$ref": "#/definitions/Link"
            },
            "type": "array"
        },
        "nodes": {
            "items": {
                "$ref": "#/definitions/NodeX"
            },
            "type": "array"
        },
        "selection": {
            "items": {
                "type": "number"
            },
            "type": "array"
        }
    },
    "type": "object",
    "required": ["links", "nodes"]
}


@pytest.mark.fast
def test_rest_core_single_node() -> None:
    """A simple single node 'hello world' test"""
    # validate schema
    Draft202012Validator.check_schema(SCHEMA)
    df2012 = Draft202012Validator(SCHEMA)
    inp_file = "single_node_helloworld.json"
    inp: Json = {}
    yaml_path = "workflow.json"
    inp_path = Path(__file__).with_name(inp_file)
    with open(inp_path, 'r', encoding='utf-8') as f:
        inp = json.load(f)
    # check if object is conformant with our schema
    df2012.is_valid(inp)
    print('----------- from rest api ----------- \n\n')
    scope = {}
    scope['type'] = 'http'

    async def receive() -> Json:
        inp_byte = json.dumps(inp).encode('utf-8')
        return {"type": "http.request", "body": inp_byte}

    # create a request object and pack it with our json payload
    req: Request = Request(scope)
    req._receive = receive
    res: Json = asyncio.run(restapi.compile_wf(req))  # call to rest api
    assert int(res['retval']) == 0



import json
from pathlib import Path
from shutil import copytree, ignore_patterns
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py as _build_py
import versioneer


class build_py(_build_py):
    """build mode script"""

    def run(self) -> None:
        adapters_dir = Path(__file__).parent / 'cwl_adapters'
        examples_dir = Path(__file__).parent / 'docs' / 'tutorials'
        adapters_target_dir = Path(self.build_lib) / 'sophios/cwl_adapters'
        examples_target_dir = Path(self.build_lib) / 'sophios/examples'
        extlist = ['*.png', '*.md', '*.rst', '*.pyc', '__pycache__', '*.json']
        copytree(adapters_dir, adapters_target_dir, dirs_exist_ok=True)
        copytree(examples_dir, examples_target_dir, ignore=ignore_patterns(*extlist), dirs_exist_ok=True)
        # Never overwrite user config
        if not (Path.home() / 'wic').exists():
            # config_path
            config_path = Path.home() / 'wic'
            # global config file in ~/wic
            global_config_file = config_path / 'global_config.json'

            # setup.py always gets executed in the project_root
            config_file = Path(__file__).parent / 'src' / 'sophios' / 'config.json'
            config = {}
            # config_file can contain absolute or relative paths
            with open(config_file, 'r', encoding='utf-8') as f:
                config = json.load(f)
            conf_tags = ['search_paths_cwl', 'search_paths_wic']
            for tag in conf_tags:
                for ns in config[tag]:
                    abs_paths = [str(Path(path).absolute()) for path in config[tag][ns]]
                    config[tag][ns] = abs_paths
            config_path.mkdir(parents=True, exist_ok=True)
            # write out the config file with paths
            with open(global_config_file, 'w', encoding='utf-8') as f:
                json.dump(config, f)

        # Continue with the standard build process
        super().run()


setup(
    name='sophios',
    version=versioneer.get_version(),
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    cmdclass={
        'build_py': build_py,
    },
    package_data={
        "sophios": ["cwl_adapters/*.cwl", "examples/*.wic"]
    }
)
